'''
Instead of using 15-bit tags and records allocated by a memory allocator (like jmalloc), to which we would point directly
using 48-bit addresses, I simplify the implementation a bit, using 32-bit keys, 32-bit values and no mem allocator,
as I am storing the keys and the values directly to the bucket entries - no need for pointing the entries to records
and then storing the values to these records.
each bucket entry has the key written in the lower bits so that i can do the lookup by key faster (no need to shift): | value | key |
'''

from mmh3 import hash  # 32-bit signed int


class _HashIndex:
    def __init__(self, n_buckets_power=4, key_len_bits=32, value_len_bits=32):
        assert n_buckets_power >= 4
        assert key_len_bits > 0
        assert value_len_bits > 0

        self.n_buckets_power = n_buckets_power
        self.n_buckets = 2**n_buckets_power
        self.key_len_bits = key_len_bits
        self.value_len_bits = value_len_bits

        self.value_mask_lower = (1 << self.value_len_bits) - 1
        self.value_mask_upper = self.value_mask_lower << self.key_len_bits
        self.key_mask_lower = (1 << self.key_len_bits) - 1
        self.bucket_mask = (1 << self.n_buckets_power) - 1

        self.bucket_len = 8
        self.buckets = [[0] * self.bucket_len for _ in range(self.n_buckets)]
        self.last_bucket = self.n_buckets - 1

        self.entries_cnt = 0
        self.capacity = 2**self.n_buckets_power * self.bucket_len
        self.resizing_threshold = int(0.75 * self.capacity)

    def _get_bucket_index(self, key):
        h = hash(key, signed=False)
        bucket_idx = h & self.bucket_mask
        assert bucket_idx < self.n_buckets
        return bucket_idx

    def _bytes_to_int(self, bytes):
        return int.from_bytes(bytes)

    def _lookup_bucket(self, bucket_idx, by_key=None, by_value=None, by_empty=False):
        for entry_idx in range(self.bucket_len - 1):
            entry = self.buckets[bucket_idx][entry_idx]
            if by_key is not None and entry & self.key_mask_lower == by_key:
                return bucket_idx, entry_idx
            if by_empty and entry == 0:
                return bucket_idx, entry_idx
            if by_value is not None and (entry >> self.key_len_bits) & self.value_mask_lower == by_value:
                return bucket_idx, entry_idx
        next_bucket = self.buckets[bucket_idx][self.bucket_len - 1]
        if next_bucket != 0:
            return self._lookup_bucket(next_bucket, by_key=by_key, by_value=by_value)
        return bucket_idx, None  # last bucket reached and no entry

    def __getitem__(self, key):
        # print(key)
        assert type(key) is bytes and len(key) <= self.key_len_bits // 8
        bucket_idx = self._get_bucket_index(key)
        key = self._bytes_to_int(key)
        bucket_idx, entry_idx = self._lookup_bucket(bucket_idx, by_key=key)
        if entry_idx is not None:
            return (self.buckets[bucket_idx][entry_idx] & self.value_mask_upper) >> self.key_len_bits
        return None

    def __setitem__(self, key, value):
        # print(key, value)
        assert type(key) is bytes and len(key) <= self.key_len_bits // 8 and type(value) is int
        bucket_idx = self._get_bucket_index(key)
        key = self._bytes_to_int(key)
        bucket_idx, entry_idx = self._lookup_bucket(bucket_idx, by_key=key)
        if entry_idx is not None:
            # if item in the buckets, update in place
            if value == 0:
                self.buckets[bucket_idx][entry_idx] = 0
                self.entries_cnt -= 1
            else:
                self.buckets[bucket_idx][entry_idx] = (value << self.key_len_bits) | key
            return
        if value == 0:  # do not insert empty values (equiv to delete)
            return
        # if not, find an empty slot and write it
        bucket_idx, entry_idx = self._lookup_bucket(bucket_idx, by_empty=True)
        if entry_idx is not None:
            self.buckets[bucket_idx][entry_idx] = (value << self.key_len_bits) | key
            self.entries_cnt += 1
            return
        # if not empty slot, allocate a new bucket and write it there
        self.buckets.append([0] * self.bucket_len)
        self.last_bucket += 1
        # update the last bucket to point to the newly added bucket
        self.buckets[bucket_idx][self.bucket_len - 1] = self.last_bucket
        # actually insert the new value
        self.buckets[self.last_bucket][0] = (value << self.key_len_bits) | key
        # increment the entries counter
        self.entries_cnt += 1
        return

    def __delitem__(self, key):
        # set value to zero
        self.__setitem__(key, 0)

    def __contains__(self, key):
        return self.__getitem__(key) is not None


class HashIndex:
    def __init__(self, n_buckets_power=4, key_len_bits=32, value_len_bits=32):
        self._hash_index = _HashIndex(n_buckets_power=n_buckets_power, key_len_bits=key_len_bits, value_len_bits=value_len_bits)

        self.n_buckets_power = n_buckets_power
        self.key_len_bits = key_len_bits
        self.value_len_bits = value_len_bits

    def __getitem__(self, key):
        return self._hash_index.__getitem__(key)

    def __setitem__(self, key, value):
        self._hash_index.__setitem__(key, value)
        # check if needs resizing
        if self._hash_index.entries_cnt > self._hash_index.resizing_threshold:
            self.resize()

    def __delitem__(self, key):
        return self._hash_index.__delitem__(key)

    def __contains__(self, key):
        return self._hash_index.__contains__(key)

    def resize(self):
        new_hash_index = _HashIndex(n_buckets_power=self.n_buckets_power + 1,
            key_len_bits=self.key_len_bits,
            value_len_bits=self.value_len_bits)

        for bucket in self._hash_index.buckets:
            for entry in bucket:
                key = entry & self._hash_index.key_mask_lower
                value = entry >> self._hash_index.key_len_bits
                if value == 0:
                    continue
                new_hash_index[key.to_bytes(self._hash_index.key_len_bits // 8)] = value

        self._hash_index = new_hash_index



import shutil
import unittest
from pathlib import Path

from fuzzytester import FuzzyTester
from kevo import AppendLog, PathRemote, MinioRemote


class TestAppendLog(unittest.TestCase, FuzzyTester):
    dir = Path('./data_test')

    def setUp(self):
        self.remote = PathRemote('/tmp/remote')
        # self.remote = MinioRemote('testbucket')
        self.dir.mkdir()

    def tearDown(self):
        shutil.rmtree(self.dir.name)
        self.remote.destroy()

    def test_basic(self):
        l = AppendLog(self.dir.name)

        l.set(b'a', b'a')
        l.set(b'asdf', b'\x00\x01\x00\x00')
        l.set(b'to be deleted', b'delete me')
        l.set(b'b', b'\x00\x00\x02\x00')
        l.set(b'd', b'3\x002\x00')
        l.set(b'a', b'a1')
        l.set(b'e', b'55')
        l.set(b'to be deleted', b'')

        self.assertEqual(l.get(b'a'), b'a1')
        self.assertEqual(l.get(b'asdf'), b'\x00\x01\x00\x00')
        self.assertEqual(l.get(b'b'), b'\x00\x00\x02\x00')
        self.assertEqual(l.get(b'c'), b'')
        self.assertEqual(l.get(b'd'), b'3\x002\x00')
        self.assertEqual(l.get(b'e'), b'55')
        self.assertEqual(l.get(b'to be deleted'), b'')

        l.close()

    def test_remote(self):
        db = AppendLog(self.dir.name, remote=self.remote)
        db.set(b'a', b'1')
        db.set(b'b', b'2')
        db.snapshot(0)
        db.set(b'a', b'3')
        db.set(b'b', b'4')
        db.snapshot(1)
        db.close()

        shutil.rmtree(self.dir.name)

        db = AppendLog(self.dir.name, remote=self.remote)
        self.assertEqual(db.get(b'a'), b'3')
        self.assertEqual(db.get(b'b'), b'4')
        db.restore(version=0)
        self.assertEqual(db.get(b'a'), b'1')
        self.assertEqual(db.get(b'b'), b'2')
        db.close()

    def test_fuzzy_generic(self):
        self.fuzzy_test(AppendLog, args={'data_dir': self.dir.name, 'compaction': True, 'remote': None},
                        key_len_range=(1, 10), val_len_range=(0, 10), n_items=1000, n_iter=1_000_000, seeds=[1],
                        test_recovery=False, test_remote=False)

    def test_fuzzy_granular(self):
        self.fuzzy_test(AppendLog, args={'data_dir': self.dir.name, 'threshold': 100, 'remote': None},
                        key_len_range=(1, 10), val_len_range=(0, 10), n_items=100, n_iter=10_000, seeds=[1],
                        test_recovery=True, test_remote=False)

    def test_fuzzy_compaction(self):
        self.fuzzy_test(AppendLog,
                        args={'data_dir': self.dir.name, 'threshold': 100, 'compaction': True, 'remote': None},
                        key_len_range=(1, 10), val_len_range=(0, 10), n_items=100, n_iter=10_000, seeds=[1],
                        test_recovery=True, test_remote=False)

    def test_fuzzy_recovery(self):
        self.fuzzy_test(AppendLog, args={'data_dir': self.dir.name, 'threshold': 1_000, 'remote': self.remote},
                        key_len_range=(1, 10), val_len_range=(0, 10), n_items=100, n_iter=10_000, seeds=[1],
                        test_recovery=True, test_remote=True)

    def test_fuzzy_large_kvs(self):
        self.fuzzy_test(AppendLog, args={'data_dir': self.dir.name, 'max_key_len': 100_000, 'max_value_len': 100_000,
                                         'threshold': 1_000, 'remote': None},
                        key_len_range=(1, 100_000), val_len_range=(0, 100_000), n_items=10, n_iter=1000, seeds=[1],
                        test_recovery=True, test_remote=False)

    def test_fuzzy_snapshot(self):
        self.fuzzy_test_snapshot(AppendLog, args={'data_dir': self.dir.name, 'threshold': 100, 'remote': self.remote},
                                 key_len_range=(1, 10), val_len_range=(0, 13), n_items=10_000, n_iter=10_000, seed=1)

    def test_fuzzy_snapshot_continuous(self):
        self.fuzzy_test_snapshot_continuous(AppendLog, args={'data_dir': self.dir.name, 'threshold': 100,
                                                             'remote': self.remote}, key_len_range=(1, 10),
                                            val_len_range=(0, 13), n_items=10_000, n_iter=10_000, seed=1)

    def test_rebuild(self):
        l1 = AppendLog(self.dir.name, threshold=10)

        l1.set(b'a', b'1')
        l1.set(b'b', b'2')
        l1.set(b'c', b'3')

        l1.close()

        l2 = AppendLog(self.dir.name)

        self.assertEqual(l2.get(b'a'), b'1')
        self.assertEqual(l2.get(b'b'), b'2')
        self.assertEqual(l2.get(b'c'), b'3')

        l2.close()


if __name__ == "__main__":
    unittest.main()



import shutil
import unittest
from pathlib import Path

from fuzzytester import FuzzyTester
from kevo import MemOnly, PathRemote, MinioRemote


class TestMemOnly(unittest.TestCase, FuzzyTester):
    dir = Path('./data_test')

    def setUp(self):
        self.remote = PathRemote('/tmp/remote')
        # self.remote = MinioRemote('testbucket')
        self.dir.mkdir()

    def tearDown(self):
        shutil.rmtree(self.dir.name)
        self.remote.destroy()

    def test_basic(self):
        db = MemOnly(self.dir)

        db.set(b'b', b'2')
        db.set(b'asdf', b'12345')
        db.set(b'cc', b'cici345')
        db.set(b'b', b'3')

        self.assertEqual(db.get(b'b'), b'3')
        self.assertEqual(db.get(b'asdf'), b'12345')
        self.assertEqual(db.get(b'cc'), b'cici345')

        db.close()

    def test_fuzzy_basic(self):
        self.fuzzy_test(MemOnly, args={'data_dir': self.dir.name, 'remote': self.remote}, key_len_range=(1, 10),
                        val_len_range=(0, 10), n_items=1_000, n_iter=10_000, seeds=[1], test_recovery=True,
                        test_remote=True)

    def test_fuzzy_snapshot_continuous(self):
        self.fuzzy_test_snapshot_continuous(MemOnly, args={'data_dir': self.dir.name, 'remote': self.remote},
                                            key_len_range=(1, 10), val_len_range=(0, 13), n_items=10_000, n_iter=10_000,
                                            seed=1)

    def test_versioning(self):
        db = MemOnly(self.dir, remote=self.remote)

        db.set(b'1', b'1')
        db.set(b'2', b'2')
        db.snapshot(0)
        db.set(b'1', b'3')
        db.snapshot(1)
        db.close()

        shutil.rmtree(self.dir)

        db = MemOnly(self.dir, remote=self.remote)

        self.assertEqual(db.get(b'1'), b'3')
        self.assertEqual(db.get(b'2'), b'2')

        db.close()


if __name__ == "__main__":
    unittest.main()



import unittest

from kevo.common import RingBuffer


class TestRingBuffer(unittest.TestCase):
    def test_basic(self):
        r = RingBuffer(5)
        self.assertTrue(r.is_empty())
        self.assertEqual(len(r), 0)
        r.add(1)
        r.add(2)
        r.add(3)
        self.assertEqual(len(r), 3)
        self.assertEqual(r.pop(), 1)
        r.add(4)
        self.assertEqual(r.pop(), 2)
        r.add(5)
        r.add(1)
        r.add(2)
        with self.assertRaises(BufferError) as ctx:
            r.add(3)
        r.pop()
        r.pop()
        r.pop()
        r.pop()
        r.pop()
        with self.assertRaises(BufferError) as ctx:
            r.pop()

        with self.assertRaises(ValueError) as ctx:
            r = RingBuffer(0)

        r = RingBuffer(1)
        r.add(1)
        self.assertEqual(len(r), 1)
        self.assertTrue(r.is_full())
        self.assertFalse(r.is_empty())
        self.assertEqual(r.pop(), 1)
        self.assertEqual(len(r), 0)
        self.assertTrue(r.is_empty())
        self.assertFalse(r.is_full())

    def test_indexing(self):
        r = RingBuffer(3)
        o1 = r.add(0)
        o2 = r.add(1)
        self.assertEqual(r[o1], 0)
        self.assertEqual(r[o2], 1)
        r.pop()
        r.pop()
        with self.assertRaises(IndexError) as ctx:
            _ = r[o1]
            _ = r[o2]

        o1 = r.add(3)
        r.add(4)
        o2 = r.add(5)
        while o1 < o2:
            r.pop()
            o1 += 1
        self.assertEqual(r.pop(), 5)
        self.assertTrue(r.is_empty())

        r.add(5)
        o = r.add(6)
        r[o] = 0
        self.assertEqual(r[o], 0)
        with self.assertRaises(IndexError) as ctx:
            _ = r[o + 1]
        self.assertEqual(r[o - 1], 5)
        with self.assertRaises(IndexError) as ctx:
            _ = r[o - 2]
        r.add(7)
        self.assertTrue(r.is_full())
    
        r = RingBuffer(1)
        o = r.add(0)
        self.assertEqual(r[o], r.pop())


if __name__ == "__main__":
    unittest.main()



import unittest

from random import Random
from kevo.common import FencePointers


class TestFencePointers(unittest.TestCase):
    def test_ser_der(self):
        rng = Random(1)
        fp1 = FencePointers()
        for i in range(100):
            randbytes = rng.randbytes(rng.randint(50, 100))
            randint = rng.randint(0, 1000)
            fp1.add(randbytes, randint)
        fp2 = FencePointers(from_str=fp1.serialize())
        self.assertEqual(fp1.pointers, fp2.pointers)


if __name__ == "__main__":
    unittest.main()



# I wrote this because pybloomfiltermmap3 does not work with python 3.11 yet
# https://en.wikipedia.org/wiki/Bloom_filter#Probability_of_false_positives.

from sys import getsizeof
import json
from base64 import b64encode, b64decode
from math import log, ceil, floor

from mmh3 import hash
from bitarray import bitarray


class BloomFilter:
    def __init__(self, est_num_items=1000, false_positive_prob=0.01, from_str: str | None = None):
        if type(from_str) is str:
            data = json.loads(from_str)

            arr = bitarray(endian=data['endian'])
            arr.frombytes(b64decode(data['bytes']))
            self.bitarray_size = data['bitarray_size']
            self.bitarray = arr[:self.bitarray_size]

            self.num_hash_funcs = data['num_hash_funcs']

            self.est_num_items = data['est_num_items']
            self.false_positive_prob = data['false_positive_prob']

        else:
            self.est_num_items = est_num_items
            self.false_positive_prob = false_positive_prob

            # https://stackoverflow.com/questions/658439/how-many-hash-functions-does-my-bloom-filter-need
            self.bitarray_size = ceil(-(est_num_items * log(false_positive_prob)) / (log(2) ** 2))
            self.num_hash_funcs = floor((self.bitarray_size / est_num_items) * log(2))  # floor to keep the hash funcs as few as possible for performance

            self.bitarray = bitarray(self.bitarray_size)
            self.bitarray.setall(False)

    def add(self, item):
        for i in range(self.num_hash_funcs):
            self.bitarray[hash(item, i) % self.bitarray_size] = True
    
    def __contains__(self, item):
        for i in range(self.num_hash_funcs):
            if not self.bitarray[hash(item, i) % self.bitarray_size]:
                return False
        return True

    def serialize(self):
        return json.dumps({
            'bytes': b64encode(self.bitarray.tobytes()).decode(),
            'bitarray_size': len(self.bitarray),
            'endian': self.bitarray.endian(),
            'num_hash_funcs': self.num_hash_funcs,
            # the last two are not necessary but I put them for completeness
            'est_num_items': self.est_num_items,
            'false_positive_prob': self.false_positive_prob
        })

    def __str__(self) -> str:
        return self.serialize()

    def __sizeof__(self):
        return getsizeof(self.bitarray)



from sys import getsizeof


class RingBuffer:
    '''
    __getitem__(), __setitem__() here are specific to my use case (to this
    codebase) and would not be present in a generic ring buffer implementation
    '''
    def __init__(self, capacity: int):
        if capacity <= 0:
            raise ValueError('capacity has to be > 0')
        self.capacity = capacity
        self.buffer = [None] * capacity
        self.write_idx = 1
        self.read_idx = 0

    def __len__(self):
        return self.write_idx - self.read_idx - 1

    def __getitem__(self, key):
        if key >= self.write_idx or key <= self.read_idx:
            raise IndexError('index out of bounds')
        return self.buffer[key % self.capacity]

    def __setitem__(self, key, value):
        if key >= self.write_idx or key <= self.read_idx:
            raise IndexError('index out of bounds')
        self.buffer[key % self.capacity] = value

    def is_empty(self):
        return self.__len__() == 0

    def is_full(self):
        return self.__len__() == self.capacity

    def add(self, element):
        if self.is_full():
            raise BufferError('buffer full')
        self.buffer[self.write_idx % self.capacity] = element
        ret = self.write_idx
        self.write_idx += 1
        return ret

    def pop(self):
        if self.is_empty():
            raise BufferError('buffer empty')
        self.read_idx += 1
        return self.buffer[self.read_idx % self.capacity]

    def set_tail_offset(self, offset: int):
        self.write_idx = offset + 1
        self.read_idx = offset

    def __sizeof__(self):
        return getsizeof(self.buffer)



import unittest
from random import Random

from kevo.common import HashIndex


class TestHashIndex(unittest.TestCase):
    def test_basic(self):
        h = HashIndex()
        h[b'asdf'] = 12
        self.assertEqual(h[b'asdf'], 12)
        del h[b'asdf']
        self.assertEqual(h[b'asdf'], None)

    def test_e2e(self):
        rng = Random(1)
        h = HashIndex()
        d = {}
        rand_keys = [rng.randint(1, 100).to_bytes(4) for _ in range(1000)]
        rand_values = [rng.randint(0, 100) for _ in range(1000)]
        for k, v in zip(rand_keys, rand_values):
            h[k] = v
            d[k] = v
            if k in d and v == 0:  # value 0 is equivalent to deletion
                del d[k]
        for k, v in zip(rand_keys, rand_values):
            if k in d:
                self.assertEqual(h[k], d[k])


if __name__ == "__main__":
    unittest.main()



from sys import getsizeof
from typing import Optional
from io import FileIO

from kevo.engines.kvstore import KVStore, Record, discover_run_files
from kevo.remote import Remote


class AppendLog(KVStore):
    name = 'AppendLog'

    def __init__(self,
                 data_dir='./data',
                 max_key_len=255,
                 max_value_len=255,
                 max_runs_per_level=3,
                 threshold=4_000_000,
                 compaction=False,
                 remote: Optional[Remote] = None):
        self.type = 'appendlog'
        super().__init__(data_dir, max_key_len=max_key_len, max_value_len=max_value_len, remote=remote)

        assert max_runs_per_level > 1
        assert threshold > 0

        self.max_runs_per_level = max_runs_per_level
        self.threshold = threshold
        self.counter = 0

        self.compaction_enabled = compaction

        self.hash_index: dict[bytes, Record] = {}
        self.levels: list[int] = []
        # read file-descriptors
        self.rfds: list[list[FileIO]] = []
        # write file-descriptor
        self.wfd: Optional[FileIO] = None

        self.global_version = 0

        if self.remote:
            self.restore()
        else:
            self.rebuild_indices()

    def rebuild_indices(self):
        self.levels.clear()
        self.rfds.clear()

        self.hash_index.clear()

        # do file discovery
        runs_discovered = self.discover_runs()
        if not runs_discovered:
            self.wfd = (self.data_dir / f'L{0}.{0}.{self.global_version}.run').open('wb')
            self.rfds.append([(self.data_dir / f'L{0}.{0}.{self.global_version}.run').open('rb')])
            self.levels.append(0)
            return

        # rebuild the index
        for level_idx, run_idx, version in sorted(runs_discovered):
            while level_idx >= len(self.levels):
                self.levels.append(0)
            self.levels[level_idx] += 1

            while level_idx >= len(self.rfds):
                self.rfds.append([])
            self.rfds[level_idx].append((self.data_dir / f'L{level_idx}.{run_idx}.{version}.run').open('rb'))

        # sort by first field asc and by second field desc
        for level_idx, run_idx, version in sorted(runs_discovered, key=lambda x: (-x[0], x[1])):
            log_file = self.rfds[level_idx][run_idx]
            offset = log_file.tell()
            k, _ = self._read_kv_pair(log_file)
            while k:
                self.hash_index[k] = Record(level_idx, run_idx, offset)
                offset = log_file.tell()
                k, _ = self._read_kv_pair(log_file)

        self.wfd = (self.data_dir / f'L{0}.{self.levels[0]}.{self.global_version}.run').open('wb')
        self.rfds[0].append((self.data_dir / f'L{0}.{self.levels[0]}.{self.global_version}.run').open('rb'))

    def close(self):
        self.close_run()
        self.wfd.close()
        for rfds in self.rfds:
            for rfd in rfds:
                rfd.close()

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def get(self, key: bytes):
        assert type(key) is bytes
        assert 0 < len(key) <= self.max_key_len

        if key not in self.hash_index:
            return KVStore.EMPTY

        record = self.hash_index[key]

        log_file = self.rfds[record.level][record.run]
        log_file.seek(record.offset)
        k, v = self._read_kv_pair(log_file)
        assert k == key
        return v

    def set(self, key: bytes, value: bytes = KVStore.EMPTY):
        assert type(key) is bytes and type(value) is bytes
        assert 0 < len(key) <= self.max_key_len and len(value) <= self.max_value_len

        if not value and key in self.hash_index:
            del self.hash_index[key]
            return

        # always write the latest log of the first level
        offset = self.wfd.tell()
        self._write_kv_pair(self.wfd, key, value, flush=True)
        self.hash_index[key] = Record(0, self.levels[0], offset)
        self.counter += len(key) + len(value)

        if self.counter >= self.threshold:
            self.close_run()
            self.open_new_files()

    def close_run(self):
        if self.counter == 0:
            return

        flush_level = 0
        if flush_level >= len(self.levels):
            self.levels.append(0)
        if flush_level >= len(self.rfds):
            self.rfds.append([])

        self.counter = 0
        self.wfd.close()

        if self.compaction_enabled:
            self.compaction(self.levels[flush_level])

        self.levels[flush_level] += 1
        if self.levels[flush_level] >= self.max_runs_per_level:
            self.merge(flush_level)

    def open_new_files(self):
        flush_level = 0
        self.wfd.close()
        self.wfd = (self.data_dir / f'L{flush_level}.{self.levels[flush_level]}.{self.global_version}.run').open('ab')
        self.rfds[flush_level].append((self.data_dir / f'L{flush_level}.{self.levels[flush_level]}.{self.global_version}.run').open('rb'))

    def compaction(self, run):
        log_path = (self.data_dir / f'L0.{run}.{self.global_version}.run')
        compacted_log_path = log_path.with_suffix('.tmp')
        # NOTE i can copy the index here and keep the old one for as long as the compaction is running to enable reads
        # concurrently

        with compacted_log_path.open('ab') as compacted_log_file:
            read_offset = 0
            self.rfds[0][run].seek(read_offset)
            k, v = self._read_kv_pair(self.rfds[0][run])
            while k:
                if k in self.hash_index and self.hash_index[k] == Record(0, run, read_offset):
                    write_offset = compacted_log_file.tell()
                    self._write_kv_pair(compacted_log_file, k, v)
                    self.hash_index[k] = Record(0, run, write_offset)
                read_offset = self.rfds[0][run].tell()
                k, v = self._read_kv_pair(self.rfds[0][run])

        self.rfds[0][run].close()
        # rename the file back
        compacted_log_path.rename(compacted_log_path.with_suffix('.run'))
        # get a new read fd
        self.rfds[0][run] = log_path.open('rb')

    def merge(self, level: int):
        next_level = level + 1
        if next_level >= len(self.levels):
            self.levels.append(0)
            self.rfds.append([])
        next_run = self.levels[next_level]

        dst_file = (self.data_dir / f'L{next_level}.{next_run}.{self.global_version}.run').open('ab')
        for run_idx in range(self.levels[level]):
            src_file = self.rfds[level][run_idx]
            src_offset = 0
            src_file.seek(src_offset)
            k, v = self._read_kv_pair(src_file)
            while k:
                if k in self.hash_index and self.hash_index[k] == Record(level, run_idx, src_offset):
                    dst_offset = dst_file.tell()
                    self._write_kv_pair(dst_file, k, v)
                    self.hash_index[k] = Record(next_level, next_run, dst_offset)
                src_offset = src_file.tell()
                k, v = self._read_kv_pair(src_file)
        dst_file.close()

        self.rfds[next_level].append((self.data_dir / f'L{next_level}.{next_run}.{self.global_version}.run').open('rb'))

        # delete merged files
        for rfd in self.rfds[level]:
            rfd.close()
        self.rfds[level].clear()

        # remove the files after successfully merging.
        for file in self.data_dir.glob(f'L{level}.*.*.run'):
            file.unlink()

        # update the runs counter
        self.levels[level] = 0
        self.levels[next_level] += 1

        # bump the global version
        self.global_version += 1

        # merge recursively
        if self.levels[next_level] >= self.max_runs_per_level:
            self.merge(next_level)

    def snapshot(self, id: int):
        self.close_run()
        if self.remote:
            runs = discover_run_files(self.data_dir)
            self.remote.push_deltas(runs, id)
        self.open_new_files()

    def restore(self, version=None):
        self.close_run()
        if self.remote:
            self.global_version = self.remote.restore(version=version)
            if self.wfd is not None:
                self.wfd.close()
            for rfds in self.rfds:
                for rfd in rfds:
                    rfd.close()
            self.rebuild_indices()

    def __sizeof__(self):
        return getsizeof(self.hash_index)



from sys import getsizeof
from typing import Optional

from kevo.engines.kvstore import KVStore, discover_run_files
from kevo.remote import Remote


class MemOnly(KVStore):
    name = 'MemOnly'

    def __init__(self,
                 data_dir='./data',
                 max_key_len=255,
                 max_value_len=255,
                 remote: Optional[Remote] = None):
        self.type = 'memonly'
        super().__init__(data_dir, max_key_len=max_key_len, max_value_len=max_value_len, remote=remote)

        self.hash_index: dict[bytes, bytes] = {}

        self.global_version = 0

        if self.remote:
            self.restore()
        else:
            self.rebuild_indices()

    def rebuild_indices(self):
        self.hash_index.clear()

        runs_discovered = self.discover_runs()
        if not runs_discovered:
            return

        level_idx, run_idx, version = runs_discovered[0]
        with (self.data_dir / f'L{level_idx}.{run_idx}.{version}.run').open('rb') as log_file:
            key, value = self._read_kv_pair(log_file)
            while key:
                self.hash_index[key] = value
                key, value = self._read_kv_pair(log_file)

    def close(self):
        pass

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def get(self, key: bytes):
        assert type(key) is bytes
        assert 0 < len(key) <= self.max_key_len

        ret = KVStore.EMPTY
        try:
            ret = self.hash_index[key]
        except KeyError:
            pass
        return ret

    def set(self, key: bytes, value: bytes = KVStore.EMPTY):
        assert type(key) is bytes and type(value) is bytes
        assert 0 < len(key) <= self.max_key_len and len(value) <= self.max_value_len

        self.hash_index[key] = value

    def flush(self):
        if len(self.hash_index) == 0:
            return

        with (self.data_dir / f'L0.0.{self.global_version}.run').open('wb') as log_file:
            for key, value in self.hash_index.items():
                self._write_kv_pair(log_file, key, value)

        for file in self.data_dir.glob(f'L0.0.*.run'):
            if int(file.name.split('.')[2]) < self.global_version:
                file.unlink()

        self.global_version += 1

    def snapshot(self, id: int):
        self.flush()
        if self.remote:
            runs = discover_run_files(self.data_dir)
            self.remote.push_deltas(runs, id)

    def restore(self, version=None):
        self.flush()
        if self.remote:
            self.global_version = self.remote.restore(version=version)
            self.rebuild_indices()

    def __sizeof__(self):
        return getsizeof(self.hash_index)



from pathlib import Path
from collections import namedtuple


Record = namedtuple('Record', ['level', 'run', 'offset'])


def bytes_needed_to_encode_len(length):
    i = 0
    while 2 ** (i * 8) <= length:
        i += 1
    return i


def discover_run_files(dir_path: Path) -> list[Path]:
    return [f for f in dir_path.glob('L*.run') if f.is_file()]


def discover_version_files(dir_path: Path) -> list[Path]:
    return [f for f in dir_path.glob('version.*.txt') if f.is_file()]


class KVStore:
    EMPTY = b''

    def __init__(self,
                 data_dir='./data',
                 max_key_len=255,
                 max_value_len=255,
                 remote=None):

        self.data_dir = Path(data_dir)
        self.data_dir.mkdir(exist_ok=True)

        self.max_key_len = max_key_len
        self.max_value_len = max_value_len

        self.key_enc_len = bytes_needed_to_encode_len(self.max_key_len)
        self.val_enc_len = bytes_needed_to_encode_len(self.max_value_len)

        self.remote = remote
        if self.remote:
            self.remote.init(self.data_dir)

    def discover_runs(self) -> list[tuple[int, int, int]]:
        runs_discovered = []
        for f in discover_run_files(self.data_dir):
            level, run, version, _ = f.name.split('.')
            level = int(level[1:])
            run = int(run)
            version = int(version)
            runs_discovered.append((level, run, version))
        # has to be sorted for rebuilding to be done in the right order
        return sorted(runs_discovered)

    def _read_kv_pair(self, fd):
        first_bytes = fd.read(self.key_enc_len)
        if not first_bytes:
            return KVStore.EMPTY, KVStore.EMPTY
        key_len = int.from_bytes(first_bytes, byteorder='little')
        key = fd.read(key_len)
        val_len = int.from_bytes(fd.read(self.val_enc_len), byteorder='little')
        value = fd.read(val_len)
        return key, value

    def _write_kv_pair(self, fd, key, value, flush=False):
        fd.write(len(key).to_bytes(self.key_enc_len, byteorder='little'))
        fd.write(key)
        fd.write(len(value).to_bytes(self.val_enc_len, byteorder='little'))
        fd.write(value)
        if flush:
            fd.flush()

    # abstract methods
    def __getitem__(self, key):
        raise NotImplementedError('')

    def __setitem__(self, key, value):
        raise NotImplementedError('')

    def get(self, key: bytes):
        raise NotImplementedError('')

    def set(self, key: bytes, value: bytes):
        raise NotImplementedError('')

    def __sizeof__(self):
        raise NotImplementedError('')

    def close(self):
        raise NotImplementedError('')



import shutil
import unittest
from pathlib import Path

from fuzzytester import FuzzyTester
from kevo import HybridLog, PathRemote, MinioRemote


class TestHybridLog(unittest.TestCase, FuzzyTester):
    dir = Path('./data_test')

    def setUp(self):
        self.remote = PathRemote('/tmp/remote')
        # self.remote = MinioRemote('testbucket')
        self.dir.mkdir()

    def tearDown(self):
        shutil.rmtree(self.dir.name)
        self.remote.destroy()

    def test_basic(self):
        l = HybridLog(self.dir.name, ro_lag_interval=1, flush_interval=1)

        l.set(b'asdf', b'\x00\x01\x00\x00')
        l.set(b'b', b'\x00\x00\x02\x00')
        l.set(b'd', b'3\x002\x00')
        l.set(b'e', b'55')

        self.assertEqual(l.get(b'asdf'), b'\x00\x01\x00\x00')
        self.assertEqual(l.get(b'b'), b'\x00\x00\x02\x00')
        self.assertEqual(l.get(b'c'), b'')
        self.assertEqual(l.get(b'd'), b'3\x002\x00')
        self.assertEqual(l.get(b'e'), b'55')

        l.close()

    def test_remote(self):
        db = HybridLog(self.dir.name, remote=self.remote)
        db.set(b'a', b'1')
        db.set(b'b', b'2')
        db.snapshot(0)
        db.set(b'a', b'3')
        db.set(b'b', b'4')
        db.snapshot(1)
        db.close()

        shutil.rmtree(self.dir.name)

        db = HybridLog(self.dir.name, remote=self.remote)
        self.assertEqual(db.get(b'a'), b'3')
        self.assertEqual(db.get(b'b'), b'4')
        db.restore(version=0)
        self.assertEqual(db.get(b'a'), b'1')
        self.assertEqual(db.get(b'b'), b'2')
        db.close()

    def test_fuzzy_realistic(self):
        self.fuzzy_test(HybridLog, args={'data_dir': self.dir.name}, key_len_range=(1, 10), val_len_range=(0, 10),
                        n_items=1000, n_iter=1_000_000, seeds=[1], test_recovery=False, test_remote=False)

    def test_fuzzy_merge(self):
        self.fuzzy_test(HybridLog, args={'data_dir': self.dir.name, 'ro_lag_interval': 10,
                                         'flush_interval': 10}, key_len_range=(1, 4), val_len_range=(0, 4),
                        n_items=1_000, n_iter=10_000, seeds=[1], test_recovery=False, test_remote=False)

    def test_fuzzy_index_rebuild(self):
        self.fuzzy_test(HybridLog, args={'data_dir': self.dir.name, 'ro_lag_interval': 10,
                                         'flush_interval': 10}, key_len_range=(1, 4), val_len_range=(0, 4),
                        n_items=1_000, n_iter=10_000, seeds=[1], test_recovery=True, test_remote=False)

    def test_fuzzy_remote(self):
        self.fuzzy_test(HybridLog, args={'data_dir': self.dir.name, 'ro_lag_interval': 10,
                                         'flush_interval': 10, 'remote': self.remote}, key_len_range=(1, 4),
                        val_len_range=(0, 4), n_items=1_000, n_iter=10_000, seeds=[1], test_recovery=True,
                        test_remote=True)

    def test_fuzzy_large_kvs(self):
        self.fuzzy_test(HybridLog, args={'data_dir': self.dir.name, 'max_key_len': 100_000, 'max_value_len': 100_000,
                                         'ro_lag_interval': 100, 'flush_interval': 100,
                                         'remote': None}, key_len_range=(1, 100_000), val_len_range=(0, 100_000),
                        n_items=10, n_iter=100, seeds=[1], test_recovery=False, test_remote=False)

    def test_fuzzy_altogether(self):
        self.fuzzy_test(HybridLog,
                        args={'data_dir': self.dir.name, 'ro_lag_interval': 150_000,
                              'flush_interval': 150_000, 'remote': self.remote},
                        key_len_range=(1, 10), val_len_range=(0, 10), n_items=10_000, n_iter=1_000_000, seeds=[1],
                        test_recovery=True, test_remote=True)

    def test_fuzzy_snapshot(self):
        self.fuzzy_test_snapshot(HybridLog,
                                 args={'data_dir': self.dir.name, 'ro_lag_interval': 400,
                                       'flush_interval': 400, 'remote': self.remote},
                                 key_len_range=(1, 10), val_len_range=(0, 13), n_items=10_000, n_iter=10_000, seed=1)

    def test_fuzzy_snapshot_continuous(self):
        self.fuzzy_test_snapshot_continuous(HybridLog, args={'data_dir': self.dir.name,
                                                             'ro_lag_interval': 400, 'flush_interval': 400,
                                                             'remote': self.remote}, key_len_range=(1, 10),
                                            val_len_range=(0, 13), n_items=10_000, n_iter=10_000, seed=1)


if __name__ == "__main__":
    unittest.main()



from collections import namedtuple
from sys import getsizeof
from typing import Optional
from io import FileIO
import struct

from sortedcontainers import SortedDict

from kevo.common import BloomFilter, FencePointers
from kevo.engines.kvstore import KVStore, discover_run_files
from kevo.remote import Remote

Run = namedtuple('Run', ['filter', 'pointers', 'nr_records'])


def append_indices(file_descriptor, fence_pointers, bloom_filter, nr_records):
    pointers_offset = file_descriptor.tell()
    file_descriptor.write(fence_pointers.serialize().encode())
    bloom_offset = file_descriptor.tell()
    file_descriptor.write(bloom_filter.serialize().encode())
    # pack two 8 byte unsigned ints for the offsets of the pointers and the bloom filter
    file_descriptor.write(struct.pack('<QQQ', pointers_offset, bloom_offset, nr_records))


class LSMTree(KVStore):
    name = 'LSMTree'

    def __init__(self,
                 data_dir='./data',
                 max_key_len=255,
                 max_value_len=255,
                 max_runs_per_level=3,
                 density_factor=20,
                 memtable_bytes_limit=1_000_000,
                 remote: Optional[Remote] = None):
        self.type = 'lsmtree'
        super().__init__(data_dir=data_dir, max_key_len=max_key_len, max_value_len=max_value_len, remote=remote)

        assert max_runs_per_level > 1
        assert density_factor > 0
        assert memtable_bytes_limit > 0

        self.max_runs_per_level = max_runs_per_level
        self.density_factor = density_factor

        self.memtable = SortedDict()
        self.memtable_bytes_limit = memtable_bytes_limit
        self.memtable_bytes_count = 0

        self.wal_path = self.data_dir / 'wal'
        if self.wal_path.is_file():
            with self.wal_path.open('rb') as wal_file:
                k, v = self._read_kv_pair(wal_file)
                while k:
                    # write the value to the memtable directly, no checks for amount of bytes etc.
                    self.memtable[k] = v
                    k, v = self._read_kv_pair(wal_file)
        self.wal_file = self.wal_path.open('ab')

        self.levels: list[list[Run]] = []
        self.rfds: list[list[FileIO]] = []

        # global version is used to attach a version number to every file flushed and merged.
        # this is used in snapshotting in the delta maps
        self.global_version = 0

        if self.remote:
            # restore calls rebuild_indices, so this way we avoid rebuilding twice
            self.restore()
        else:
            self.rebuild_indices()

    def rebuild_indices(self):
        # TODO set the global version to the max of the discovered ones
        self.levels.clear()
        self.rfds.clear()

        # do file discovery
        runs_discovered = self.discover_runs()
        if not runs_discovered:
            return

        # load filters and pointers for levels and runs
        for level_idx, run_idx, version in sorted(runs_discovered):
            with (self.data_dir / f'L{level_idx}.{run_idx}.{version}.run').open('rb') as run_file:
                # fetch the last 24 bytes (=3*8)
                run_file.seek(-24, 2)
                bloom_end_offset = run_file.tell()
                offsets = run_file.read()
                pointers_offset, bloom_offset, nr_records = struct.unpack('<QQQ', offsets)
                run_file.seek(pointers_offset)
                pointers = FencePointers(from_str=run_file.read(bloom_offset - pointers_offset).decode())
                run_file.seek(bloom_offset)
                bloom_filter = BloomFilter(from_str=run_file.read(bloom_end_offset - bloom_offset).decode())

            while level_idx >= len(self.levels):
                self.levels.append([])
            self.levels[level_idx].append(Run(bloom_filter, pointers, nr_records))

            while level_idx >= len(self.rfds):
                self.rfds.append([])
            self.rfds[level_idx].append((self.data_dir / f'L{level_idx}.{run_idx}.{version}.run').open('rb'))

    def close(self):
        self.wal_file.close()
        for rfds in self.rfds:
            for rfd in rfds:
                rfd.close()

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def get(self, key: bytes):
        assert type(key) is bytes
        assert 0 < len(key) <= self.max_key_len

        if key in self.memtable:
            return self.memtable[key]

        for level_idx, level in enumerate(self.levels):
            for i, run in reversed(list(enumerate(level))):
                if key in run.filter:
                    # bisect -1 because I want the index of the item on the left
                    idx = run.pointers.bisect(key) - 1
                    if idx < 0:
                        idx = 0
                    _, offset = run.pointers.peekitem(idx)
                    run_file = self.rfds[level_idx][i]
                    run_file.seek(offset)
                    for _ in range(run.pointers.density_factor):
                        read_key, read_value = self._read_kv_pair(run_file)
                        if read_key == key:
                            return read_value

        return KVStore.EMPTY

    def set(self, key: bytes, value: bytes = KVStore.EMPTY):
        assert type(key) is bytes and type(value) is bytes
        assert 0 < len(key) <= self.max_key_len and len(value) <= self.max_value_len

        if key not in self.memtable:
            self.memtable_bytes_count += len(key) + len(value)

        # NOTE maybe i should write after the flush?
        # cause this way the limit is not a hard limit, it may be passed by up to 255 bytes
        self.memtable[key] = value

        if self.memtable_bytes_count > self.memtable_bytes_limit:
            # normally I would allocate a new memtable here so that writes can continue there
            # and then give the flushing of the old memtable to a background thread
            self.flush()
        else:
            # write to wal
            self._write_kv_pair(self.wal_file, key, value)

    def merge(self, level_idx: int):
        level = self.levels[level_idx]
        if level_idx + 1 >= len(self.levels):
            self.levels.append([])
        next_level = self.levels[level_idx + 1]

        fence_pointers = FencePointers(self.density_factor)
        # I can replace with an actual accurate count but I don't think it's worth it, it's an estimate anyway
        bloom_filter = BloomFilter(sum([run.filter.est_num_items for run in level]))

        # use counters and the nr_records in the file to know when the key-values part has been parsed
        # the alternative would be to use tell() all the time which is slow
        fds, keys, values, is_empty, counters, nr_records_in_run = [], [], [], [], [], []
        for i, _ in enumerate(level):
            fd = self.rfds[level_idx][i]
            fd.seek(0)
            fds.append(fd)
            k, v = self._read_kv_pair(fd)
            keys.append(k)
            values.append(v)
            is_empty.append(True if not k else False)
            counters.append(0 if not k else 1)
            nr_records_in_run.append(self.levels[level_idx][i].nr_records)

        nr_records = 0
        with (self.data_dir / f'L{level_idx + 1}.{len(next_level)}.{self.global_version}.run').open('wb') as run_file:
            while not all(is_empty):
                argmin_key = len(level) - 1
                # correctly initialize the argmin_key (cause empty key b'' would make it instantly the argmin_key in
                # the next for loop which we don't want)
                for i in reversed(range(len(level))):
                    if not is_empty[i]:
                        argmin_key = i
                        break
                for i in reversed(range(len(level))):
                    if not is_empty[i] and keys[i] < keys[argmin_key]:
                        argmin_key = i

                # assumption: empty value == deleted item, so if empty I am writing nothing
                if values[argmin_key]:
                    fence_pointers.add(keys[argmin_key], run_file.tell())
                    self._write_kv_pair(run_file, keys[argmin_key], values[argmin_key])
                    bloom_filter.add(keys[argmin_key])
                    nr_records += 1

                written_key = keys[argmin_key]

                # read next kv pair
                if not is_empty[argmin_key]:
                    if counters[argmin_key] >= nr_records_in_run[argmin_key]:
                        is_empty[argmin_key] = True
                    else:
                        keys[argmin_key], values[argmin_key] = self._read_kv_pair(fds[argmin_key])
                        counters[argmin_key] += 1

                # skip duplicates
                # + 1 cause inclusive range
                for i in reversed(range(argmin_key + 1)):
                    # if it's the same key, read one more pair to skip it
                    while not is_empty[i] and written_key == keys[i]:
                        if counters[i] >= nr_records_in_run[i]:
                            is_empty[i] = True
                        else:
                            keys[i], values[i] = self._read_kv_pair(fds[i])
                            counters[i] += 1

            append_indices(run_file, fence_pointers, bloom_filter, nr_records)

        if level_idx + 1 >= len(self.rfds):
            self.rfds.append([])
        self.rfds[level_idx + 1].append((self.data_dir / f'L{level_idx + 1}.{len(next_level)}.{self.global_version}.run').open('rb'))
        for fd in fds:
            fd.close()
        self.rfds[level_idx].clear()

        self.global_version += 1

        # remove the files after successfully merging.
        for file in self.data_dir.glob(f'L{level_idx}.*.*.run'):
            file.unlink()

        # empty the runs array
        level.clear()

        # append new run
        next_level.append(Run(bloom_filter, fence_pointers, nr_records))

        # cascade the merging recursively
        if len(next_level) >= self.max_runs_per_level:
            self.merge(level_idx + 1)

    def flush(self):
        if len(self.memtable) == 0:
            return
        fence_pointers = FencePointers(self.density_factor)
        bloom_filter = BloomFilter(len(self.memtable))

        flush_level = 0  # always flush at first level

        if flush_level >= len(self.levels):
            self.levels.append([])
        if flush_level >= len(self.rfds):
            self.rfds.append([])

        n_runs = len(self.levels[0])

        nr_records = 0
        with (self.data_dir / f'L{flush_level}.{n_runs}.{self.global_version}.run').open('wb') as run_file:
            while self.memtable:
                k, v = self.memtable.popitem(0)
                fence_pointers.add(k, run_file.tell())
                self._write_kv_pair(run_file, k, v)
                bloom_filter.add(k)
                nr_records += 1
            append_indices(run_file, fence_pointers, bloom_filter, nr_records)

        self.memtable_bytes_count = 0

        self.levels[flush_level].append(Run(bloom_filter, fence_pointers, nr_records))
        self.rfds[flush_level].append((self.data_dir / f'L{flush_level}.{n_runs}.{self.global_version}.run').open('rb'))

        # reset WAL
        self.wal_file.close()
        self.wal_file = self.wal_path.open('wb')

        # trigger merge if exceeding the runs per level
        if len(self.levels[flush_level]) >= self.max_runs_per_level:
            self.merge(flush_level)

    def snapshot(self, id: int):
        self.flush()
        if self.remote:
            runs = discover_run_files(self.data_dir)
            self.remote.push_deltas(runs, id)

    def restore(self, version=None):
        # flush first to empty the memtable
        self.flush()
        if self.remote:
            self.global_version = self.remote.restore(version=version)
            # close open file descriptors first
            for rfds in self.rfds:
                for rfd in rfds:
                    rfd.close()
            self.rebuild_indices()

    def __sizeof__(self):
        memtable_size = sum((getsizeof(k) + getsizeof(v) for k, v in self.memtable.items()))
        bloom_filters_size = sum((getsizeof(run.filter) for level in self.levels for run in level))
        fence_pointers_size = sum((getsizeof(run.pointers) for level in self.levels for run in level))
        return memtable_size + bloom_filters_size + fence_pointers_size



from .lsmtree import LSMTree
from .hybridlog import HybridLog
from .appendlog import AppendLog
from .memonly import MemOnly

__all__ = [
    'LSMTree',
    'HybridLog',
    'AppendLog',
    'MemOnly'
]



import os
import shutil
from time import sleep
from pathlib import Path
from typing import Optional

from minio import Minio, S3Error

from kevo.engines.kvstore import discover_run_files, discover_version_files


def read_key(filename):
    with open(filename, 'r') as f:
        return f.read()


# abstract class
class Remote:
    def __init__(self):
        self.src_dir_path = None

    def init(self, src_dir_path):
        self.src_dir_path = src_dir_path

    def push_deltas(self, delta_map: list[Path]):
        raise NotImplementedError

    def put(self, filename: str):
        raise NotImplementedError

    def get(self, filename: str):
        raise NotImplementedError

    def gc(self):
        raise NotImplementedError

    def restore(self, version=None):
        raise NotImplementedError

    def destroy(self):
        raise NotImplementedError


class PathRemote(Remote):
    def __init__(self, remote_dir_path: str, latency_per_byte: int = 0):
        # latency_per_byte simulates network delay, useful for benchmarking and tests
        super().__init__()

        self.remote_dir_path = Path(remote_dir_path)
        self.latency_per_byte = latency_per_byte

        os.makedirs(self.remote_dir_path, exist_ok=True)

    def init(self, src_dir_path: Path):
        super().init(src_dir_path)

        os.makedirs(self.remote_dir_path, exist_ok=True)

    def push_deltas(self, delta_map: list[Path], snapshot_version: int = 0):
        delta_map = {f.name for f in delta_map}

        discovered_files = discover_run_files(self.remote_dir_path)
        remote_delta_map = {f.name for f in discovered_files}

        files_to_push = delta_map - remote_delta_map
        for file in files_to_push:
            self.put(file)

        version_file_name = f'version.{snapshot_version}.txt'
        with open(os.path.join(self.src_dir_path.resolve(), version_file_name), 'w') as f:
            f.write('-'.join(delta_map))
        self.put(version_file_name)

    def _simulate_net_delay(self, filepath: str):
        filesize = os.path.getsize(filepath)
        if self.latency_per_byte > 0:
            sleep(self.latency_per_byte * filesize)

    def put(self, filename: str):
        filename = os.path.basename(filename)
        filepath = os.path.join(self.src_dir_path.resolve(), filename)
        self._simulate_net_delay(filepath)
        shutil.copy(filepath, self.remote_dir_path)

    def get(self, filename: str):
        filename = os.path.basename(filename)
        shutil.copy(os.path.join(self.remote_dir_path, filename), self.src_dir_path.resolve())

    def gc(self):
        raise NotImplementedError

    def restore(self, version=None):
        if version is None:
            # find the latest
            discovered_files = discover_version_files(self.remote_dir_path)
            if not discovered_files:
                return 0
            version = max(map(lambda p: int(p.name.split('.')[1]), discovered_files))

        # clean up the local tree first
        # TODO do this after a successful recovery, fetch all the files first and only afterwards delete the old ones.
        shutil.rmtree(self.src_dir_path)
        os.mkdir(self.src_dir_path)

        version_file_name = f'version.{version}.txt'
        self.get(version_file_name)
        with open(os.path.join(self.src_dir_path, version_file_name), 'r') as f:
            filenames = f.read().split('-')

        for f in filenames:
            self.get(f)

        global_version = 0
        global_versions = [int(f.name.split('.')[2]) for f in self.remote_dir_path.glob('L*.run') if f.is_file()]
        if global_versions:
            global_version = max(global_versions) + 1
        return global_version

    def destroy(self):
        shutil.rmtree(self.remote_dir_path)


class MinioRemote(Remote):
    def __init__(self,
                 bucket: str,
                 address='localhost:9000',
                 access_key_fname='access.key',
                 secret_key_fname='secret.key',
                 minio_client=None):
        super().__init__()

        self.bucket = bucket
        self.client = minio_client if minio_client else Minio(address, read_key(access_key_fname),
                                                              read_key(secret_key_fname), secure=False)

        self.delta_map: Optional[set[str]] = None

        if not self.client.bucket_exists(self.bucket):
            self.client.make_bucket(self.bucket)

        self.global_version = 0

    def init(self, src_dir_path: Path):
        super().init(src_dir_path)

        if not self.client.bucket_exists(self.bucket):
            self.client.make_bucket(self.bucket)

        self.delta_map = {obj.object_name for obj in self.client.list_objects(self.bucket) if
                          obj.object_name.startswith('L') and obj.object_name.endswith('run')}

    # TODO move this to Remote class
    def push_deltas(self, delta_map: list[Path], snapshot_version: int = 0):
        delta_map = {f.name for f in delta_map}

        files_to_push = delta_map - self.delta_map
        for file in files_to_push:
            self.put(file)

        version_file_name = f'version.{snapshot_version}.txt'
        with open(os.path.join(self.src_dir_path.resolve(), version_file_name), 'w') as f:
            f.write('-'.join(delta_map))
        self.put(version_file_name)

    def put(self, filename: str):
        filename = os.path.basename(filename)
        filepath = os.path.join(self.src_dir_path.resolve(), filename)
        self.client.fput_object(self.bucket, filename, filepath)

    def get(self, filename: str):
        filename = os.path.basename(filename)
        self.client.fget_object(self.bucket, filename, os.path.join(self.src_dir_path.resolve(), filename))

    def gc(self):
        raise NotImplementedError

    def restore(self, version=None):
        if version is None:
            # find the latest
            discovered_files = {obj.object_name for obj in self.client.list_objects(self.bucket) if
                                obj.object_name.startswith('version') and obj.object_name.endswith('txt')}
            if not discovered_files:
                return 0
            version = max(map(lambda p: int(p.split('.')[1]), discovered_files))

        # clean up the local tree first
        # TODO do this after a successful recovery, fetch all the files first and only afterwards delete the old ones.
        shutil.rmtree(self.src_dir_path)
        os.mkdir(self.src_dir_path)

        version_file_name = f'version.{version}.txt'
        self.get(version_file_name)
        with open(os.path.join(self.src_dir_path, version_file_name), 'r') as f:
            filenames = f.read().split('-')

        for f in filenames:
            self.get(f)

        global_version = 0
        global_versions = [int(obj.object_name.split('.')[2]) for obj in self.client.list_objects(self.bucket) if
                                obj.object_name.startswith('L') and obj.object_name.endswith('run')]
        if global_versions:
            global_version = max(global_versions) + 1
        return global_version

    def destroy(self):
        for o in self.client.list_objects(self.bucket):
            self.client.remove_object(self.bucket, o.object_name)
        self.client.remove_bucket(self.bucket)



from io import FileIO
from sys import getsizeof
from typing import Optional

# from kevo.common.hashindex import HashIndex
from kevo.common.ringbuffer import RingBuffer
from kevo.engines.kvstore import KVStore, Record, discover_run_files
from kevo.remote import Remote


class HybridLog(KVStore):
    name = 'HybridLog'

    # TODO make the mem_segment_len equal to the sum of
    # ro_lag_int + flush_int
    def __init__(self,
                 data_dir='./data',
                 max_key_len=255,
                 max_value_len=255,
                 max_runs_per_level=3,
                 ro_lag_interval=2 ** 10,
                 flush_interval=(4 * 2 ** 10),
                 hash_index='dict',
                 remote: Optional[Remote] = None):
        self.type = 'hybridlog'
        super().__init__(data_dir, max_key_len=max_key_len, max_value_len=max_value_len, remote=remote)

        assert max_runs_per_level > 1
        assert flush_interval > 0
        assert ro_lag_interval > 0
        assert hash_index in ['dict', 'native'], 'hash_index parameter must be either "dict" or "native"'

        self.max_runs_per_level = max_runs_per_level
        self.ro_lag_interval = ro_lag_interval
        self.flush_interval = flush_interval
        # The mem segment length is not a parameter because there is no reason using more memory than the required
        # amount, there is no trade-off. Therefore we keep it to the lowest possible value which is this sum.
        self.mem_segment_len = flush_interval + ro_lag_interval + 1

        if hash_index == 'native':
            # TODO
            # self.hash_index = HashIndex(n_buckets_power=4, key_len_bits=self.max_key_len*8, value_len_bits=self.max_value_len*8)
            raise NotImplementedError("do not use the native hash index, it currently has a bug, use 'dict' instead")
        else:
            self.hash_index: dict[bytes, int] = {}

        self.la_to_file_offset: dict[int, Record] = {}

        self.head_offset: int = 0  # LA > head_offset is in mem
        self.ro_offset: int = 0  # in LA > ro_offset we have the mutable region
        self.tail_offset: int = 0  # points to the tail of the log, the last record inserted

        self.levels: list[int] = []
        # read file-descriptors
        self.rfds: list[list[FileIO]] = []
        # write file-descriptor
        self.wfd: Optional[FileIO] = None

        self.global_version = 0

        self.memory: Optional[RingBuffer] = None

        if self.remote:
            self.restore()
        else:
            self.rebuild_indices()

    def rebuild_indices(self):
        self.levels.clear()
        self.rfds.clear()

        self.hash_index.clear()
        self.la_to_file_offset.clear()

        self.head_offset = 0
        self.ro_offset = 0
        self.tail_offset = 0

        self.memory = RingBuffer(self.mem_segment_len)

        # do file discovery
        runs_discovered = self.discover_runs()
        if not runs_discovered:
            self.wfd = (self.data_dir / f'L{0}.{0}.{self.global_version}.run').open('wb')
            self.rfds.append([(self.data_dir / f'L{0}.{0}.{self.global_version}.run').open('rb')])
            return

        # rebuild the index
        for level_idx, run_idx, version in sorted(runs_discovered):
            while level_idx >= len(self.levels):
                self.levels.append(0)
            self.levels[level_idx] += 1

            while level_idx >= len(self.rfds):
                self.rfds.append([])
            self.rfds[level_idx].append((self.data_dir / f'L{level_idx}.{run_idx}.{version}.run').open('rb'))

        # sort by first field asc and by second field desc
        for level_idx, run_idx, version in sorted(runs_discovered, key=lambda x: (-x[0], x[1])):
            log_file = self.rfds[level_idx][run_idx]
            offset = log_file.tell()
            k, _ = self._read_kv_pair(log_file)
            while k:
                self.head_offset += 1
                self.hash_index[k] = self.head_offset
                self.la_to_file_offset[self.head_offset] = Record(level_idx, run_idx, offset)
                offset = log_file.tell()
                k, _ = self._read_kv_pair(log_file)

        self.wfd = (self.data_dir / f'L{0}.{self.levels[0]}.{self.global_version}.run').open('wb')
        self.rfds[0].append((self.data_dir / f'L{0}.{self.levels[0]}.{self.global_version}.run').open('rb'))

        self.ro_offset = self.head_offset
        self.tail_offset = self.ro_offset
        self.memory.set_tail_offset(self.tail_offset)

    def close(self):
        self.flush(self.tail_offset)  # flush everything
        self.wfd.close()
        for rfds in self.rfds:
            for rfd in rfds:
                rfd.close()

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def get(self, key: bytes):
        assert type(key) is bytes
        assert 0 < len(key) <= self.max_key_len

        if key not in self.hash_index:
            return KVStore.EMPTY

        offset = self.hash_index[key]
        if offset > self.head_offset:
            _, v = self.memory[offset]
            return v

        record = self.la_to_file_offset[offset]
        log_file = self.rfds[record.level][record.run]
        log_file.seek(record.offset)
        _, v = self._read_kv_pair(log_file)
        return v

    def set(self, key: bytes, value: bytes = KVStore.EMPTY):
        assert type(key) is bytes and type(value) is bytes
        assert 0 < len(key) <= self.max_key_len and len(value) <= self.max_value_len

        if self.memory.is_full():
            self.flush(self.ro_offset)

        if key in self.hash_index:
            offset = self.hash_index[key]
            if offset > self.ro_offset:
                # update in-place
                self.memory[offset] = (key, value)
                return

        self.tail_offset = self.memory.add((key, value))
        self.hash_index[key] = self.tail_offset
        # no need to increment the tail offset as the ring buffer returns the new (incremented) address

        if self.tail_offset - self.ro_offset > self.ro_lag_interval:
            self.ro_offset += 1

        if self.ro_offset - self.head_offset > self.flush_interval:
            self.flush(self.ro_offset)
            self.open_new_files()

    def flush(self, offset: int):
        if self.memory.is_empty():
            return

        flush_level = 0

        if flush_level >= len(self.levels):
            self.levels.append(0)
        if flush_level >= len(self.rfds):
            self.rfds.append([])

        while self.head_offset < offset:
            key, value = self.memory.pop()
            write_offset = self.wfd.tell()
            self.head_offset += 1
            # if is not the most recent record, drop it, no need to keep it.
            if self.hash_index[key] == self.head_offset:
                self._write_kv_pair(self.wfd, key, value)
                self.la_to_file_offset[self.head_offset] = Record(flush_level, self.levels[flush_level], write_offset)

        self.wfd.close()

        self.levels[flush_level] += 1
        if self.levels[flush_level] >= self.max_runs_per_level:
            self.merge(flush_level)

        # open a new file after merging
        # self.wfd = (self.data_dir / f'L{flush_level}.{self.levels[flush_level]}.{self.global_version}.run').open('ab')
        # self.rfds[flush_level].append((self.data_dir / f'L{flush_level}.{self.levels[flush_level]}.{self.global_version}.run').open('rb'))

    def merge(self, level: int):
        next_level = level + 1
        if next_level >= len(self.levels):
            self.levels.append(0)
            self.rfds.append([])
        next_run = self.levels[next_level]

        dst_file = (self.data_dir / f'L{next_level}.{next_run}.{self.global_version}.run').open('ab')
        for run_idx in range(self.levels[level]):
            src_file = self.rfds[level][run_idx]
            src_offset = 0
            src_file.seek(src_offset)
            k, v = self._read_kv_pair(src_file)
            while k:
                if k in self.hash_index:
                    la = self.hash_index[k]
                    if la in self.la_to_file_offset and self.la_to_file_offset[la] == Record(level, run_idx, src_offset):
                        dst_offset = dst_file.tell()
                        self._write_kv_pair(dst_file, k, v)
                        self.la_to_file_offset[la] = Record(next_level, next_run, dst_offset)
                src_offset = src_file.tell()
                k, v = self._read_kv_pair(src_file)
        dst_file.close()

        self.rfds[next_level].append((self.data_dir / f'L{next_level}.{next_run}.{self.global_version}.run').open('rb'))

        # delete merged files
        for rfd in self.rfds[level]:
            rfd.close()
        self.rfds[level].clear()
        # remove the files after successfully merging.
        for file in self.data_dir.glob(f'L{level}.*.*.run'):
            file.unlink()

        # update the runs counter
        self.levels[level] = 0
        self.levels[next_level] += 1

        # bump the global version
        self.global_version += 1

        # merge recursively
        if self.levels[next_level] >= self.max_runs_per_level:
            self.merge(next_level)

    def open_new_files(self):
        flush_level = 0
        self.wfd.close()
        self.wfd = (self.data_dir / f'L{flush_level}.{self.levels[flush_level]}.{self.global_version}.run').open('ab')
        self.rfds[flush_level].append((self.data_dir / f'L{flush_level}.{self.levels[flush_level]}.{self.global_version}.run').open('rb'))

    def snapshot(self, id: int):
        self.flush(self.tail_offset)
        self.ro_offset = self.tail_offset
        if self.remote:
            runs = discover_run_files(self.data_dir)
            self.remote.push_deltas(runs, id)
        self.open_new_files()

    def restore(self, version=None):
        if self.remote:
            self.global_version = self.remote.restore(version=version)
            # close open file descriptors before rebuilding
            if self.wfd is not None:
                self.wfd.close()
            for rfds in self.rfds:
                for rfd in rfds:
                    rfd.close()
            self.rebuild_indices()

    def __sizeof__(self):
        return getsizeof(self.hash_index) + getsizeof(self.la_to_file_offset) + getsizeof(self.memory)



import unittest
from random import Random

from kevo.common import BloomFilter


class TestBloomFilter(unittest.TestCase):
    def test_false_positive_probability(self):
        for accuracy in [3, 4, 5]:
            skip = 50
            n = 1000000
            fp_prob = 10 ** (-accuracy)
            b = BloomFilter(int(n / skip), fp_prob)

            for i in range(n):
                s = (str(i)).encode()
                if i % skip == 0:
                    b.add(s)

            fp_count = 0
            for i in range(n):
                s = (str(i)).encode()
                self.assertFalse(i % skip == 0 and s not in b)
                if i % skip != 0 and s in b:
                    fp_count += 1

            self.assertAlmostEqual(fp_count / n, fp_prob, 4)
        
    def test_ser_der(self):
        rng = Random(1)
        b1 = BloomFilter()
        for i in range(20):
            b1.add(rng.randbytes(10))
        b2 = BloomFilter(from_str=b1.serialize())
        self.assertEqual(b1.bitarray, b2.bitarray)


if __name__ == "__main__":
    unittest.main()



import setuptools
classifiers = [
    "Programming Language :: Python :: 3",
]
setuptools.setup(
    name="kevo",
    version="0.0.0",
    author="Nikos Gavalas",
    description="Key-value store with 3 backend engines",
    classifiers=classifiers,
    keywords=["key-value", "store"],
    packages=setuptools.find_packages(),
    install_requires=[
        'sortedcontainers >= 2.4.0, < 3',
        'bitarray >= 2.6.2, < 3',
        'mmh3 >= 3.0.0, < 4',
        'minio >= 7.1.14, < 8',
    ],
    python_requires='>=3.11',
)



'''
This now is implemented as a sorted dictionary (because I need the bisect_left/right) with base64/json-based ser/der.
A better implementation would be: two arrays (one for keys one for values) so that I can binary-search on the keys, and binary encoding for ser/der.
TODO rebuilding from string could be done linearly if the serialization is sorted, right now the sorteddict is being rebuilt from scratch so that should be fixed
'''

from sys import getsizeof
import json
from base64 import b64encode, b64decode

from sortedcontainers import SortedDict


class FencePointers:
    def __init__(self, density_factor=20, from_str: str | None = None):
        self.pointers = SortedDict()

        self.density_factor = density_factor
        self.counter = 0

        if type(from_str) is str:
            data = json.loads(from_str) 
            for k, v in data['pointers'].items():
                self.pointers[b64decode(k)] = v
            self.density_factor = data['density_factor']
            self.counter = data['counter']

    def add(self, key: bytes, offset: int):
        if self.counter % self.density_factor == 0:
            self.pointers[key] = offset

        self.counter += 1
    
    def bisect(self, key: bytes):
        return self.pointers.bisect(key)
    
    def peekitem(self, idx):
        return self.pointers.peekitem(idx)

    def serialize(self):
        pointers = {}
        for k, v in self.pointers.items():
            pointers[b64encode(k).decode()] = v
        return json.dumps({
            'pointers': pointers,
            'density_factor': self.density_factor,
            'counter': self.counter
        })

    def __len__(self):
        return self.counter

    def __str__(self) -> str:
        return self.serialize()

    def __sizeof__(self):
        return sum((getsizeof(k) + getsizeof(v) for k, v in self.pointers.items()))
